/* $Id: ncbi_localnet.c,v 1.1.1.1 2010/05/19 18:36:00 rhubley Exp $
 * ===========================================================================
 *
 *                            PUBLIC DOMAIN NOTICE
 *               National Center for Biotechnology Information
 *
 *  This software/database is a "United States Government Work" under the
 *  terms of the United States Copyright Act.  It was written as part of
 *  the author's official duties as a United States Government employee and
 *  thus cannot be copyrighted.  This software/database is freely available
 *  to the public for use. The National Library of Medicine and the U.S.
 *  Government have not placed any restriction on its use or reproduction.
 *
 *  Although all reasonable efforts have been taken to ensure the accuracy
 *  and reliability of the software and data, the NLM and the U.S.
 *  Government do not and cannot warrant the performance or results that
 *  may be obtained by using this software or data. The NLM and the U.S.
 *  Government disclaim all warranties, express or implied, including
 *  warranties of performance, merchantability or fitness for any particular
 *  purpose.
 *
 *  Please cite the author in any work or product based on this material.
 *
 * ===========================================================================
 *
 * Author:  Anton Lavrentiev
 *
 * File Description:
 *   Get IP address of CGI client and determine the IP locality
 *
 *   NOTE:  This is an internal NCBI-only API not for export to third parties.
 *
 */

#include "../ncbi_ansi_ext.h"
#include "../ncbi_priv.h"
#include <connect/ncbi_connutil.h>
#if defined(NCBI_OS_UNIX)
#  include <connect/ext/ncbi_ifconf.h>
#elif !defined(INADDR_LOOPBACK)
#  define INADDR_LOOPBACK 0x1F000001
#endif /*NCBI_OS_UNIX*/
#include <connect/ext/ncbi_crypt.h>
#include <connect/ext/ncbi_localnet.h>
#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef NCBI_OS_UNIX
#  include <sys/param.h>
#endif /*NCBI_OS_UNIX*/

#define NCBI_USE_ERRCODE_X   Connect_LocalNet

#if defined(NCBI_OS_MSWIN) && !defined(PATH_MAX)
#  define PATH_MAX _MAX_PATH
#endif


typedef enum {
    eLocalIP_None = 0,
    eLocalIP_Host,
    eLocalIP_Range,
    eLocalIP_Network
} ELocalIPType;


typedef struct {
    ELocalIPType type;
    unsigned int a, b;
} SLocalIP;


static SLocalIP s_LocalIP[256 + 1] = { { eLocalIP_None } };


static SLocalIP kLocalIP[] = {
    /* localnet (localhost) */
    { eLocalIP_Network, INADDR_LOOPBACK-1, 0xFFFFFF00 }, /* 127.0.0.xxx      */
    /* from assigned IP ranges */
    { eLocalIP_Range,   0x820E0800, 0x820E09FF }, /* 130.14.{8|9}.0/24       */
    { eLocalIP_Range,   0x820E0B00, 0x820E0CFF }, /* 130.14.1{1|2}.0/24      */
    { eLocalIP_Range,   0x820E1400, 0x820E1AFF }, /* 130.14.20.0..27.255     */
    { eLocalIP_Range,   0x820E1B40, 0x820E1BFF }, /*  w/o 130.14.27.0/26     */
    { eLocalIP_Network, 0x820E1D00, 0xFFFFFF00 }, /* 130.14.29.yyy           */
    { eLocalIP_Network, 0xA5700700, 0xFFFFFF00 }, /* 165.112.7.zzz (colo)    */
    /* from private IP ranges */
    { eLocalIP_Network, 0x0A0A0000, 0xFFFF0000 }, /* 10.10/16      from cl.A */
    { eLocalIP_Network, 0x0A140000, 0xFFFF0000 }, /* 10.20/16      from cl.A */
    { eLocalIP_Network, 0xAC100000, 0xFFF00000 }, /* 172.16/12  16 nets cl.B */
    { eLocalIP_None }
};


static int/*bool*/ s_WithinLocalIP(unsigned int    addr/*host byte order*/,
                                   const SLocalIP* local)
{
    assert(local  &&  local->type != eLocalIP_None);
    switch (local->type) {
    case eLocalIP_Host:
        assert(!local->b);
        return !(addr ^ local->a);
    case eLocalIP_Range:
        assert(local->a < local->b);
        return local->a <= addr  &&  addr <= local->b;
    case eLocalIP_Network:
        assert(local->a  &&  local->b);
        return !((addr & local->b) ^ local->a);
    default:
        assert(0);
        break;
    }
    return 0/*false*/;
}


static SLocalIP s_RangeOfLocalIP(const SLocalIP* local)
{
    SLocalIP retval = { eLocalIP_Range };
    switch (local->type) {
    case eLocalIP_Host:
        retval.a =  local->a;
        retval.b =  local->a;
        break;
    case eLocalIP_Range:
        retval.a =  local->a;
        retval.b =  local->b;
        break;
    case eLocalIP_Network:
        retval.a =  local->a;
        retval.b = (local->a & local->b) | ~local->b;
        break;
    default:
        assert(0);
        /*FALLTHRU*/
    case eLocalIP_None:
        retval.a =  0;
        retval.b =  0;
        break;
    }
    return retval;
}


#if defined(_DEBUG)  &&  !defined(NDEBUG)
static void s_DumpLocalIP(const SLocalIP* local)
{
    SLocalIP range = s_RangeOfLocalIP(local);
    char addr[2][40];
    const char* what;

    switch (local->type) {
    case eLocalIP_None:
        what = "None";
        break;
    case eLocalIP_Host:
        what = "Host";
        break;
    case eLocalIP_Range:
        what = "Range";
        break;
    case eLocalIP_Network:
        what = "Network";
        break;
    default:
        assert(0);
        return;
    }
    if (SOCK_ntoa(SOCK_HostToNetLong(range.a), addr[0], sizeof(addr[0])) != 0)
        strcpy(addr[0], "?");
    if (SOCK_ntoa(SOCK_HostToNetLong(range.b), addr[1], sizeof(addr[1])) != 0)
        strcpy(addr[1], "?");
    CORE_LOGF_X(1, eLOG_Trace, ("%s %s-%s", what, addr[0], addr[1]));
}
#endif /*_DEBUG && !NDEBUG*/


static int/*bool*/ s_ParseLocalIP(SLocalIP* local, const char* s)
{
    if (!SOCK_isip(s)) {
        int dots = 0;
        local->type = eLocalIP_Host;
        for (;;) {
            char small[4];
            char* e;
            long d;

            errno = 0;
            d = (int) strtol(s, &e, 10);
            if (s == e  ||  s + 3 < e  ||  d < 0  ||  d > 255  ||  errno)
                break;
            sprintf(small, "%u", (unsigned int) d);
            if (strlen(small) != (size_t)(e - s))
                break;
            s = e;
            switch (local->type) {
            case eLocalIP_Host:
                local->a <<= 8;
                local->a  |= d;
                if (*s != '.') {
                    local->a <<= (3 - dots) << 3;
                    switch (*s) {
                    case '/':
                        local->type = eLocalIP_Network;
                        break;
                    case '-':
                        local->type = eLocalIP_Range;
                        s++;
                        continue;
                    default:
                        goto out;
                    }
                } else if (++dots <= 3) {
                    s++;
                    continue;
                } else
                    goto out;
                assert(*s == '/'  &&  local->type == eLocalIP_Network);
                if (!SOCK_isipEx(++s, 1/*fullquad*/))
                    continue;
                local->b  = SOCK_NetToHostLong(SOCK_gethostbyname(s));
                return local->a  &&  ~local->b  &&  !(local->a & ~local->b)
                    &&  !(~local->b & (~local->b + 1));
            case eLocalIP_Range:
                if (*s)
                    goto out;
                local->b  = dots > 0 ? local->a : 0;
                local->b &= ~((1 << ((4 - dots) << 3)) - 1);
                local->b |=    d << ((3 - dots) << 3);
                local->b |=   (1 << ((3 - dots) << 3)) - 1;
                return local->a < local->b;
            case eLocalIP_Network:
                if (*s  ||  d == 0  ||  d > 31)
                    return 0/*slashes are not allowed in hostnames, failure*/;
                local->b = ~((1 << (32 - d)) - 1);
                return local->a  &&  !(local->a & ~local->b);
            default:
                assert(0);
                return 0/*failure*/;
            }
        }
    out:
        /* last resort (and maybe expensive one): try as a regular host name */
        ;
    }
    if (!(local->a = SOCK_gethostbyname(s)))
        return 0/*failure*/;
    local->type = eLocalIP_Host;
    local->a    = SOCK_NetToHostLong(local->a);
    local->b    = 0;
    return 1/*success*/;
}


static int/*bool*/ s_OverlapLocalIP(const SLocalIP* local, size_t n)
{
    SLocalIP r1 = s_RangeOfLocalIP(local);
    size_t i;

    for (i = 0;  i < n;  i++) {
        SLocalIP r2    = s_RangeOfLocalIP(s_LocalIP + i);
        unsigned int a = r1.a > r2.a ? r1.a : r2.a;
        unsigned int b = r1.b < r2.b ? r1.b : r2.b;
        if (a < b)
            return 1/*true*/;
    }
    return 0/*false*/;
}


static void s_LoadLocalIPs(void)
{
    char buf[PATH_MAX + 1];
    const char* filename = ConnNetInfo_GetValue(0, "LOCAL_IPS",
                                                buf, sizeof(buf) - 1,
                                                "/etc/ncbi/local_ips");
    int lineno;
    size_t n;
    FILE *fp;

    if (filename  &&  strcasecmp(filename, "--HARDCODED--") == 0)
        filename = 0;
    if (!filename  ||  !(fp = fopen(filename, "r"))) {
        if (filename) {
            CORE_LOGF_ERRNO_X(2, errno != ENOENT ? eLOG_Error : eLOG_Trace,
                              errno, ("Cannot load local IPs from `%s'", buf));
        }
        CORE_LOG_X(3, eLOG_Trace, "Using default local IPs");
        assert(sizeof(s_LocalIP) >= sizeof(kLocalIP));
        memcpy(s_LocalIP, kLocalIP, sizeof(kLocalIP));
        return;
    }

    CORE_LOGF_X(4, eLOG_Trace, ("Loading local IPs from `%s'", filename));
    memcpy(s_LocalIP, kLocalIP, sizeof(kLocalIP[0]));
    n = 1/*localhost gets always added*/;
    lineno = 0;
    do {
        SLocalIP local;
        size_t len;
        char* c;
        if (!fgets(buf, sizeof(buf) - 1, fp))
            break;
        lineno++;
        if (!(len = strcspn(buf, "!#")))
            continue;
        if (buf[len]) {
            buf[len] = '\0';
        } else if (buf[len - 1] == '\n') {
            if (len > 1  &&  buf[len - 2] == '\r')
                len--;
            buf[len - 1] = '\0';
        }
        if (!*(c = buf + strspn(buf, " \t")))
            continue;
        len = strcspn(c, " \t");
        if (c[len]) {
            size_t tail = len + strspn(c + len, " \t");
            if (c[tail]) {
                CORE_LOGF_X(5, eLOG_Error,
                            ("Skipping noisy IP spec at line %u: `%s'",
                             lineno, c));
                continue;
            }
            c[len] = '\0';
        }
        if (!s_ParseLocalIP(&local, c)) {
            CORE_LOGF_X(6, eLOG_Error,
                        ("Skipping invalid IP spec at line %u: `%s'",
                         lineno, c));
            continue;
        }
        if (n >= sizeof(s_LocalIP)/sizeof(s_LocalIP[0])) {
            CORE_LOGF_X(7, eLOG_Error,
                        ("Too many local IP specs, max %u allowed",
                         (unsigned int)(n - 1)));
            break;
        }
        if (s_OverlapLocalIP(&local, n)) {
            CORE_LOGF_X(8, eLOG_Warning,
                        ("Local IP spec at line %u `%s' overlaps with"
                         " already defined one(s)", lineno, c));
        }
        s_LocalIP[n++] = local;
    } while (!feof(fp));
    fclose(fp);

    CORE_LOGF_X(9, eLOG_Trace, ("Done loading local IPs, %u line%s, %u entr%s",
                                lineno, &"s"[lineno == 1], (unsigned int) n,
                                n == 1 ? "y" : "ies"));
    if (n < sizeof(s_LocalIP)/sizeof(s_LocalIP[0]))
        s_LocalIP[n].type = eLocalIP_None;
}


extern int/*bool*/ NcbiIsLocalIP(unsigned int ip)
{
    static int/*bool*/ s_Inited = 0/*false*/;
    unsigned int addr = SOCK_NetToHostLong(ip);
    size_t n;

    if (!s_Inited) {
        CORE_LOCK_WRITE;
        if (!s_Inited) {
            s_LoadLocalIPs();
            s_Inited = 1;
            CORE_UNLOCK;
#if defined(_DEBUG)  &&  !defined(NDEBUG)
            for (n = 0;  n < sizeof(s_LocalIP)/sizeof(s_LocalIP[0]);  n++) {
                s_DumpLocalIP(s_LocalIP + n);
                if (s_LocalIP[n].type == eLocalIP_None)
                    break;
            }
#endif /*_DEBUG && !NDEBUG*/
        } else
            CORE_UNLOCK;
    }
    for (n = 0;  n < sizeof(s_LocalIP)/sizeof(s_LocalIP[0]);  n++) {
        if (s_LocalIP[n].type == eLocalIP_None)
            break;
        else if (s_WithinLocalIP(addr, s_LocalIP + n))
            return 1/*true*/;
    }
    return 0/*false*/;
}


#ifdef __GNUC__
inline
#endif /*__GNUC__*/
static int/*bool*/ s_IsPrivateIP(unsigned int ip)
{
    unsigned int addr = SOCK_NetToHostLong(ip);
    return
        !((addr & 0xFFFFFF00) ^ (INADDR_LOOPBACK-1))  ||  /* 127.0.0.x */
        /* private [non-routable] IP ranges, according to IANA and RFC1918 */
        !((addr & 0xFF000000) ^ 0x0A000000)  || /* 10/8                      */
        !((addr & 0xFFFF0000) ^ 0xA9FE0000)  || /* 169.254/16                */
        !((addr & 0xFFF00000) ^ 0xAC100000)  || /* 172.16.0.0-172.31.255.255 */
        !((addr & 0xFFFF0000) ^ 0xC0A80000)  || /* 192.168/16                */
        /* multicast IP range is also excluded: 224.0.0.0-239.255.255.255 */
#if   defined(IN_MULTICAST)
        IN_MULTICAST(addr)
#elif defined(IN_CLASSD)
        IN_CLASSD(addr)
#else
        !((addr & 0xF0000000) ^ 0xE0000000)
#endif /*IN_MULTICAST*/
        ;
}


static const char* s_SearchTrackingEnv(const char*        name,
                                       const char* const* tracking_env)
{
    const char* const* str;
    const char* result;
    size_t name_len;

    if ( !tracking_env ) {
        result = getenv(name);
#if 0/*defined(_DEBUG) && !defined(NDEBUG)*/
        CORE_LOGF_X(10, eLOG_Note, ("Getenv(`%s') = %s%s%s", name,
                                    result ? "\""   : "",
                                    result ? result : "NULL",
                                    result ? "\""   : ""));
#endif /*_DEBUG && !NDEBUG*/
    } else {
        result = 0;
        name_len = strlen(name);
        for (str = tracking_env;  *str;  ++str) {
            if (strncasecmp(*str, name, name_len) == 0
                &&  (*str)[name_len] == '=') {
                result = &(*str)[name_len + 1];
                break;
            }
        }
#if 0/*defined(_DEBUG) && !defined(NDEBUG)*/
        CORE_LOGF_X(11, eLOG_Note, ("Tracking `%s' = %s%s%s", name,
                                    result ? "\""   : "",
                                    result ? result : "NULL",
                                    result ? "\""   : ""));
#endif /*_DEBUG && !NDEBUG*/
    }
    return result;
}


static const char* s_GetForwardedFor(const char* const* tracking_env,
                                     unsigned int* addr)
{
    const char* f = s_SearchTrackingEnv("HTTP_X_FORWARDED_FOR", tracking_env);
    int/*bool*/ external;
    char *p, *q, *r, *s;
    unsigned int ip;

    if (!f  ||  !*f)
        return 0;
    r = 0;
    external = !(s = strdup(f + strspn(f, " \t"))) ? 1 : 0;
    for (p = s;  p  &&  *p;  p += strspn(p, ", \t")) {
        q = p + strcspn(p, ", \t");
        if (*q) {
            *q++ = '\0';
        }
        if (*p  &&  (ip = SOCK_gethostbyname(p)) != 0) {
            int/*bool*/ private_ip;
            if (!(private_ip = s_IsPrivateIP(ip))  &&  !NcbiIsLocalIP(ip)) {
                r     = p;
                *addr = ip;
                break;
            } else if (!r  ||  (!private_ip  &&  s_IsPrivateIP(*addr))) {
                r     = p;
                *addr = ip;
            }
        } else {
            /* unresolvable name causes external mapping */
            external = 1;
        }
        p = q;
    }
    if (r) {
        memmove(s, r, strlen(r) + 1);
        return s;
    }
    return external ? "" : 0;
}


/* Environment checked here must be in correspondence with
 * tracking environment created in CTrackingEnvHolder::CTrackingEnvHolder()
 * (header: <cgi/ncbicgi.hpp>, source: cgi/ncbicgi.cpp, library: cgi)
 */
extern unsigned int NcbiGetCgiClientIPEx(ECgiClientIP       flag,
                                         char*              buf,
                                         size_t             buf_size,
                                         const char* const* tracking_env)
{
    struct {
        const char*  host;
        unsigned int ip;
    } probe[4];
    const char* forwarded_for = 0;
    int/*bool*/ external = 0;
    const char* host = 0;
    unsigned int ip = 0;
    size_t i;

    memset(probe, 0, sizeof(probe));
    for (i = 0;  i < sizeof(probe)/sizeof(probe[0]);  i++) {
        switch (i) {
        case 0:
            probe[i].host = s_SearchTrackingEnv("HTTP_CAF_PROXIED_HOST",
                                                tracking_env);
            break;
        case 1:
            probe[i].host = forwarded_for = s_GetForwardedFor(tracking_env,
                                                              &probe[i].ip);
            break;
        case 2:
            probe[i].host = s_SearchTrackingEnv("PROXIED_IP", tracking_env);
            break;
        case 3:
            probe[i].host = s_SearchTrackingEnv("HTTP_X_FWD_IP_ADDR",
                                                tracking_env);
            break;
        default:
            assert(0);
            continue;
        }
        if (!probe[i].host) {
            continue;
        }
        if (!probe[i].ip  &&  *probe[i].host) {
            probe[i].ip = SOCK_gethostbyname(probe[i].host);
        }
        if (NcbiIsLocalIP(probe[i].ip)) {
            continue;
        }
        external = 1;
        if (!s_IsPrivateIP(probe[i].ip)) {
            host = probe[i].host;
            ip   = probe[i].ip;
            break;
        }
    }
    if (!ip) {
        for (i = external  ||  flag == eCgiClientIP_TryLeast;  i < 8;  i++) {
            unsigned int xip = 0;
            const char* xhost;
            switch (i) {
            case 0:
                assert(!external);
                xhost = s_SearchTrackingEnv("HTTP_CLIENT_HOST", tracking_env);
                break;
            case 1:
            case 2:
            case 3:
            case 4:
                xhost = probe[i - 1].host;
                xip   = probe[i - 1].ip;
                break;
            case 5:
                xhost = s_SearchTrackingEnv("REMOTE_HOST", tracking_env);
                break;
            case 6:
                xhost = s_SearchTrackingEnv("REMOTE_ADDR", tracking_env);
                break;
            case 7:
                if (flag != eCgiClientIP_TryAll) {
                    continue;
                }
                xhost = s_SearchTrackingEnv("NI_CLIENT_IPADDR", tracking_env);
                break;
            default:
                assert(0);
                continue;
            }
            if (!xhost) {
                continue;
            }
            if (i < 1  ||  i > 4) {
                xip = *xhost ? SOCK_gethostbyname(xhost) : 0;
            }
            if (!xip) {
                continue;
            }
            if (!external  ||  (!NcbiIsLocalIP(xip)  &&  !s_IsPrivateIP(xip))){
                host = xhost;
                ip   = xip;
                break;
            }
        }
    }
    if (buf  &&  buf_size) {
        if (host) {
            strncpy0(buf, host, buf_size - 1);
        } else {
            buf[0] = '\0';
        }
    }
    if (forwarded_for  &&  *forwarded_for) {
        free((void*) forwarded_for);
    }
    return ip;
}


extern unsigned int NcbiGetCgiClientIP(ECgiClientIP       flag,
                                       const char* const* tracking_env)
{
    return NcbiGetCgiClientIPEx(flag, 0, 0, tracking_env);
}


extern int/*bool*/ NcbiIsLocalCgiClient(const char* const* tracking_env)
{
    return NcbiIsLocalIP(NcbiGetCgiClientIP(eCgiClientIP_TryAll,tracking_env));
}
